#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "f0.h"

void callback(
    float *f0p,
    float *vuvp,
    float *rms_speech,
    float *acpkp,
    int vecsize,
    void *userdata
    )
{
    int i = vecsize - 1;
    for (; i >= 0; --i) {
        printf("%f %f %f %f\n", f0p[i], vuvp[i], rms_speech[i], acpkp[i]);
    }
}

int main(int argc, char **argv)
{
  int opterr = 0,  c;
  int debug = 0;
  float beg = 0.0, end = -1.0;
  char *filename=NULL;

    get_f0_session *session = init_get_f0();
    /* defaults */
    session->par->min_f0 = 60.0;
    session->par->max_f0 = 650.0;
    session->par->wind_dur = 0.01;
    session->par->frame_step = 0.005;

  while ((c = getopt (argc, argv, "i:s:e:m:x:w:f:d")) != -1){
       switch (c)
           {
	   case 'i':
	     filename = optarg;
	     break;
	   case 's':
	     beg = atof(optarg);
	     break;
           case 'e':
	     end = atof(optarg);
	     break;
	   case 'm':
	     session->par->min_f0 = atof(optarg);
	     break;
	   case 'x':
	     session->par->max_f0 = atof(optarg);
	     break;
	   case 'w':
	     session->par->wind_dur = atof(optarg);
	     break;
	   case 'f':
	     session->par->frame_step = 0.005;
	     break;
           case 'd':
	     debug=1;
	     break;
	   default:
	     opterr = 1;
    }
  }
  if(filename==NULL)opterr=1;
  if (opterr){
      printf("Usage: get_f0a -i file [-s START] [-e END] [-m MINF0] [-x MAXF0] [-w WINDUR] [-f FRAMESTEP]\n");
      exit(1);
  }
  if (debug){
     fprintf(stderr, "min_f0 %g max_f0 %g wind_dur %g frame_step %g\n", 
	     session->par->min_f0, session->par->max_f0, session->par->wind_dur, session->par->frame_step);
  }
    get_f0(filename, session, beg, end, &callback, 0);
    close_get_f0(session);
    return 0;
}
